#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "alignment.h"

void print_usage_message(char *argv0) {
	fprintf(stderr, "%s: [-g gap_score] [-a match_score] " \
			"[-i mismatch_score] [-n] sequence_1 sequence2\n", argv0);
};

void print_help_message(char *argv0) {
	print_usage_message(argv0);
	fprintf(stderr, " -g, gap_score n\t aligns with a gap score of n " \
			"(should be negative to act as a penalty)\n");
	fprintf(stderr, " -a, match_score n\t aligns with a match score of n " \
			"(should be positive to encourage matching)\n");
	fprintf(stderr, " -i, mismatch_score n\t aligns with a mismatch score of n\n");
	fprintf(stderr, " -n, no header\t\t does not print initial line with "\
			"counts of mismatches, insertions, and deletes\n");
	fprintf(stderr, "Default scores are gap: -1, match: 5, mismatch: -1\n");
};

/**
 * Aligns the supplied sequences according to a modified Needleman-
 * Wunsch algorithm (for DNA sequences) using the supplied scoring scheme
 * gap indicates the score for generating gaps, match for generating
 * matches, and mismatch for generating mismatches
 * Note that since these are scores, they must be made negative to
 * act as a penalty, eg) gap should be negative to penalize forming gaps
 * Returns an alignment struct with the resulting aligned sequences as
 * well as counts of mismatches, insertions, and deletions of the second
 * sequence with respect to the first sequence
 */
alignment* align(char* seq1, char* seq2, int gap, int match, int mismatch) {
	int s1_len = strlen(seq1);
	int s2_len = strlen(seq2);
	
	//begin by calculating the scores
	int **scores = calloc(s1_len+1, sizeof(int*));
	int **path = calloc(s1_len+1, sizeof(int*));
	int up=1, left=2, diag=3;
	int y, x;
	for (y = 0; y < s1_len+1; y++) {
		scores[y] = calloc(s2_len+1, sizeof(int));
		path[y] = calloc(s2_len+1, sizeof(int));
	}
	
	//initialize all initial runs with gap penalty
	for (y = 0; y < s1_len; y++) {
		scores[y][0] = y * gap;
		path[y][0] = up;
	}
	for (x = 0; x < s2_len; x++) {
		scores[0][x] = x * gap;
		path[0][x] = left;
	}
	
	int mod, gap1, gap2;
	for (y = 1; y < s1_len+1; y++) {
		for (x = 1; x < s2_len+1; x++) {
			mod = scores[y-1][x-1] + ((seq1[y-1] == seq2[x-1]) ? match : mismatch);
			gap1 = scores[y-1][x] + gap;
			gap2 = scores[y][x-1] + gap;
			//take the max of all these above scores
 			if (gap2 > mod && gap2 > gap1) {
				scores[y][x] = gap2;
				path[y][x] = up;
			} else if (gap1 > mod && gap1 >= gap2) {
				scores[y][x] = gap1;
				path[y][x] = left;
			} else if (mod >= gap1 && mod >= gap2) {
				scores[y][x] = mod;
				path[y][x] = diag;
			} else {
				fprintf(stderr, "didn't find max:\n\t");
				fprintf(stderr, "%d\t%d\t%d\n", mod, gap1, gap2);
			}
		}
	}
	int score = scores[s1_len][s2_len];
	
	//scores are calculated, now align them
	int buff_len = 1024;
	char buff1[buff_len];
	buff1[buff_len-1] = '\0';
	char buff2[buff_len];
	buff2[buff_len-1] = '\0';
	int r1 = buff_len-2;
	int r2 = buff_len-2;
	y = s1_len;
	x = s2_len;
	int mismatches=0, inserts=0, deletes=0;
	while (y > 0 && x > 0) {
		if (path[y][x] == diag) {
			buff1[r1--] = seq1[(y--)-1];
			buff2[r2--] = seq2[(x--)-1];
			if (seq1[y] != seq2[x]) {
				mismatches++;
			}
		} else if (path[y][x] == up) {
			buff1[r1--] = '-';
			buff2[r2--] = seq2[(x--)-1];
			deletes++;
		} else if (path[y][x] == left) {
			buff1[r1--] = seq1[(y--)-1];
			buff2[r2--] = '-';
			inserts++;
		}
	}
	while (y > 0) {
		buff1[r1--] = seq1[(y--)-1];
		buff2[r2--] = '-';
		inserts++;
	}
	while (x > 0) {
		buff1[r1--] = '-';
		buff2[r2--] = seq2[(x--)-1];
		deletes++;
	}
	
	//generate the alignment struct, and return it
	char *res1 = calloc((y = strlen(&buff1[++r1]))+1, sizeof(char));
	char *res2 = calloc((x = strlen(&buff2[++r2]))+1, sizeof(char));
	strcpy(res1, &buff1[r1]);
	res1[y] = '\0';
	strcpy(res2, &buff2[r2]);
	res2[x] = '\0';
	
	alignment *ret = malloc(sizeof(alignment));
	ret->seq1 = res1;
	ret->seq2 = res2;
	ret->mismatches = mismatches;
	ret->inserts = inserts;
	ret->deletes = deletes;
	ret->score = score;
	
	for (y = 0; y < s1_len+1; y++) {
		free(scores[y]);
		free(path[y]);
	}
	free(scores);
	free(path);
	
	return ret;
};

/**
 * Frees up any memory allocated to an alignment struct
 */
void alignment_destroy(alignment* a) {
	if (a == NULL) {
		return;
	}
	if (a->seq1) {
		free(a->seq1);
	}
	if (a->seq2) {
		free(a->seq2);
	}
	free(a);
};

/**
 * Entry point for the standalone aligning program
 * Takes options for gap, match, and mismatch scores by command line as
 * well as the two sequences to align
 * Prints the resulting alignment struct along with information of the
 * mismatches, insertions, and deletions as an optional header
 * Default scoring scheme is
 * 	gap: -10, match: 5, mismatch: 0
 */
/*
int main(int argc, char **argv) {
	if (argc < 3) {
		if (argc > 1 && (strcmp(argv[1], "help") == 0 || strcmp(argv[1], "?"))) {
			print_help_message(argv[0]);
		} else {
			print_usage_message(argv[0]);
		}
		exit(1);
	}
	
	int gap = -1;
	int match = 5;
	int mismatch = -1;
	int header = 1;
	
	int c;
	while ((c = getopt(argc, argv, "g:a:i:n")) != EOF) {
		switch(c) {
		case 'g':
			gap = atoi(optarg);
			break;
		case 'a':
			match = atoi(optarg);
			break;
		case 'i':
			mismatch = atoi(optarg);
			break;
		case 'n':
			header = 0;
			break;
		default:
			fprintf(stderr, "%s: Unknown parameter (%s)\n", argv[0], optarg);
			exit(1);
		}
	}
	
	if (optind < argc - 2) {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	alignment* a = align(argv[optind], argv[optind+1], gap, match, mismatch);
	if (header) {
		printf("#mismatches: %d\tinserts: %d\tdeletes: %d\n",
				a->mismatches, a->inserts, a->deletes);
	}
	printf("%s\n%s\n", a->seq1, a->seq2);
	
	alignment_destroy(a);
	return 0;
};
*/
