#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "include.h"
#include "hash.h"
#include "alignment.h"

static FILE* f_in = NULL;
static FILE* f_out = NULL;
static FILE* f_stats = NULL;
static FILE* f_align = NULL;
static int single_flag = 0;
static int align_score_cutoff = 80;
static int align_top_n = 5;
static int min_unclustered_count = 1;

static hash* h = NULL;
static entry* entries = NULL;
static int num_entries = 0;
static hash *index_link = NULL;
static list **lists = NULL;
static int *mm1_counts = NULL;
static int *mm2_counts = NULL;
static int *totals = NULL;
static int *mm1_totals = NULL;
static int *mm2_totals = NULL;
static char *used = NULL;
static char *core = NULL;

static void print_usage_message(char *argv0) {
	fprintf(stderr, "usage: %s [-s stats_file] [-a align_file] [-f] " \
			"[-c cutoff] [-n k] [-m min] [-o output_file] [input_file]\n", argv0);
};

static void print_help_message(char *argv0) {
	printf("usage: %s [-s stats_file] [-a align_file] [-f] " \
			"[-c cutoff] [-n k] [-m min] [-o output_file] [input_file]\n", argv0);
	printf("input_file gives a list of sequences with respective counts " \
			"to be clustered; if no input_file specified, reads from " \
			"stdin\n\n");
	printf("  -s FILE\t calculates additional statistics " \
			"and outputs them to FILE\n");
	printf("  -a FILE\t stores alignment results in FILE\n");
	printf("  -f\t\t forces entries to be " \
			"clustered to, at most, one core\n");
	printf("  -c N\t\t only performs alignments if align score " \
			"is at least N\n");
	printf("  -n N\t\t aligns the top N candidates " \
			"during the aligning phase; defaults to 0 to skip aligning\n");
	printf("  -m N\t\t a minimum of N counts will be required " \
			"for a sequence to become a core\n");
	printf("  -o FILE\t additionally outputs results to FILE\n");
};

/**
 * Populates the hash table with count entries from the count file
 * Count file can be created with barseq count
 * The input file MUST BE non-null before calling this
 */
static void build_table() {
	char buffer[1024];
	char *ptr;
	int count;
	
	if (h == NULL) {
		h = hash_create();
	}
	
	while (fgets(buffer, 1024, f_in)) {
		//check to see if line is commented with #
		if (buffer[0] == '#') {
			continue;
		}
		
		//in format key index
		ptr = strchr(buffer, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (sequence\tcount)\n");
			continue;
		}
		
		*ptr = '\0';
		count = atoi(ptr+1);
		if (count <= 0) {
			fprintf(stderr, "Warning: an entry has a non-positive" \
				" count associated with it: %s\t%d\n", buffer, count);
		}
		//sequence is at buffer currently
		
		hash_create_entry(h, buffer, count);
	}
};

/**
 * Populates the entries variable with all the counts from the hash
 * table after the hash table has been built
 * This allows scanning through all the entries in order from highest
 * count to lowest count
 */
static void extract_entries() {
	if (h) {
		entries = hash_sorted_entries(h);
	}
};

/**
 * Calculates the number of first degeneracy mismatches of a
 * supplied sequence
 * That is, given a sequence, each base is varied and checked for
 * containment in the hash table (and thus, the count file); if it is
 * then there is a mismatch potential
 * The number of these potential mismatches is the mismatch count
 * 
 * The second parameter is a pointer to a non-null list (already
 * initialized) so that the mismatch potential sequences can be
 * recorded
 * This means that the length of the list is equal to the returned
 * mismatch count
 * 
 * Returns an integer representing mismatch count of supplied sequence
 */
static int overlap(char *seq, list *l) {
	//finds the number of entries that are almost equal
	// to supplied sequence (off by 1 bp)
	
	int len = strlen(seq);
	char *degenerated = malloc((len+1)*sizeof(char));
	
	strcpy(degenerated,seq);
	degenerated[len] = '\0';
	
	int ov = 0;
	
	int c;
	int i, j;
	for (i = 0; i < len; i++) {
		for (j = 0; j < 5; j++) {
			if (seq[i] != "ACGTN"[j]) {
				degenerated[i] = "ACGTN"[j];
				if ((c = hash_lookup(h, degenerated)) >= 0) {
					list_create_entry(l, degenerated, c);
					ov++;
				}
			}
		}
		
		degenerated[i] = seq[i];	//reset before moving on
	}
	
	free(degenerated);
	
	return ov;
}

/**
 * Checks if the supplied input sequences are within n degeneracy
 * levels of each other
 * It simply compares the best case common strand between the two
 * sequences, determining how many differences it has
 * If the differences are greater than the degeneracies, there is no
 * way of reconciling the two strands
 */
static int in_n_degen(char *s1, char *s2, int n) {
	char *i=s1, *j=s2;
	int differences = 0;
	while (*i && *j) {
		if (*i != *j) {
			differences++;
		}
		if (differences > n) {
			//no way they are in two degeneracies if they
			// disagree at 3 or more bases
			return 0;
		}
		i++;
		j++;
	}
	if (*i || *j) {
		//length mismatch
		return 0;
	}
	
	return 1;
}

static void align_general() {
	if (align_top_n == 0 || f_align == NULL) {
		//user requested to skip alignment
		return;
	}
	
	fprintf(stderr, "Aligning remaining sequences...\n");
	
	int i, j, k, l;
	int p = num_entries/10;
	alignment* al;
	int unaligned=0, aligned=0;
	
	int nmer = 4;
	int len = 1;
	int word_num, slot;
	for (i = 0; i < nmer; i++) {	//5^nmers
		len *= 5;
	}
	char **matches = calloc(len, sizeof(char*));
	for (i = 0; i < len; i++) {
		matches[i] = NULL;
	}
	
	for (i = 0; i < num_entries; i++) {
		if (core[i]) {
			word_num = strlen(entries[i].key) - nmer + 1;
			for (j = 0; j < word_num; j++) {
				slot = hash_slot_nmer(&entries[i].key[j], nmer);
				if (matches[slot] == NULL) {
					matches[slot] = calloc(num_entries, sizeof(char));
					for (k = 0; k < num_entries; k++) {
						matches[slot][k] = 0;
					}
				}
				//matches[slot][i]++;
				matches[slot][i] = 1;
			}
		}
	}
	
	char* seq_match = calloc(num_entries, sizeof(char));
	int* top_match = calloc(align_top_n, sizeof(int));
	int* top_index = calloc(align_top_n, sizeof(int));
	int* best_align_score = calloc(num_entries, sizeof(int));
	char** best_align_seq = calloc(num_entries, sizeof(char*));
	
	int average_best_align_score = 0;
	for (i = 0; i < num_entries; i++) {
		if (!used[i]) {
			for (j = 0; j < num_entries; j++) {
				seq_match[j] = 0;
			}
			word_num = strlen(entries[i].key) - nmer + 1;
			for (j = 0; j < word_num; j++) {
				slot = hash_slot_nmer(&entries[i].key[j], nmer);
				if (matches[slot]) {
					for (k = 0; k < num_entries; k++) {
						if (core[k]) {
							seq_match[k] += matches[slot][k];
						}
					}
				}
			}
			//all the core 4mer matches have been found
			for (j = 0; j < align_top_n; j++) {
				top_match[j] = 0;
				top_index[j] = -1;
			}
			for (j = 0; j < num_entries; j++) {
				if (core[j]) {
					for (k = align_top_n-1; k >= 0; k--) {
						if (seq_match[j] > top_match[k]) {
							break;
						}
					}
					if (k < 0) {
						continue;
					}
					for (l = 0; l < k; l++) {
						top_match[l] = top_match[l+1];
						top_index[l] = top_index[l+1];
					}
					top_match[k] = seq_match[j];
					top_index[k] = j;
				}
			}
			best_align_score[i] = 0;
			best_align_seq[i] = NULL;
			for (j = 0; j < align_top_n; j++) {
				if (top_index[j] >= 0) {
					al = align(entries[j].key, entries[i].key, -1, 5, -1);
					if (al->score > best_align_score[i]) {
						best_align_score[i] = al->score;
						best_align_seq[i] = entries[j].key;
					}
					alignment_destroy(al);
				}
			}
			if (best_align_score[i] >= align_score_cutoff && best_align_seq[i]) {
				aligned++;
			} else {
				unaligned++;
			}
			average_best_align_score += best_align_score[i];
		}
		if (i % p == 0) {
			fprintf(stderr, "\t%d of %d entries [%d%%]\n", i,
					num_entries, (i*10)/p);
		}
	}
	average_best_align_score = average_best_align_score / (aligned + unaligned);
	
	int ij;
	if (f_align) {
		fprintf(f_align, "#Alignment score cutoff: %d\tAligned top %d " \
				"candidates\n", align_score_cutoff, align_top_n);
		fprintf(f_align, "#Aligned sequences: %d\tUnaligned sequences: %d\n",
				aligned, unaligned);
		fprintf(f_align, "#Average best-alignment score: %d\n",
				average_best_align_score);
		fprintf(f_align, "#Core sequence\tCore count\t" \
				"Aligning sequence\tAlign seq count\tScore\n");
		for (i = 0; i < num_entries; i++) {
			if (best_align_score[i] >= align_score_cutoff && best_align_seq[i]) {
				ij = hash_lookup(index_link, best_align_seq[i]);
				fprintf(f_align, "%s\t%d\t%s\t%d\t%d\n", entries[ij].key,
						entries[ij].value, entries[i].key,
						entries[i].value, best_align_score[i]);
			}
		}
	}
	
	for (i = 0; i < len; i++) {
		if (matches[i]) {
			free(matches[i]);
		}
	}
	free(matches);
	free(top_match);
	free(top_index);
	free(best_align_score);
	free(best_align_seq);
}

/**
 * Entry point for the cluster function
 * There are four main phases to the clustering process
 * 	1. File input and entry building:
 * The file is read, hash table populated, and entries created
 * 	2. Finding the mismatch counts and lists
 * For each of the entries, find the mismatch counts and generate the
 * lists; the overlap function does the work in this part
 * 	3. Create clusters
 * For each entry (in order from highest count to lowest), take the
 * group of mismatches for that entry
 * In that group, the group entry with highest mismatch count is selected
 * to become the core of the cluster
 * Once the core is selected, the list of mismatches is linked to the core
 * and each of the mismatches are removed from the pool of available
 * sequences for further clustering
 * There are two possibilities for the count of the cluster then, if
 * the single flag is given, the counts of a clustering sequences are
 * added completely to the core
 * Without the flag, the counts of the clustering sequences are divided
 * up between the potential cores it can join, weighted by the count
 * of the core sequence
 * 	4. Align unclustered sequence
 * Attempt to distribute the unclustered sequences to cores based on
 * how close it is in mismatches, and then how well they align
 * 
 * If the statistics flag is given, there will be additional statistics
 * provided in the intermediate steps
 * Statistics are generated for the clustering possibilities of each
 * sequence in the count file
 * These statistics include the proximity of clustering (1 or 2
 * degeneracies), the sequence name, the number of clusters that it
 * may belong to, and the sizes of the clusters
 */
int cluster_main(int argc, char **argv) {
	if (argc > 1 && (strstr(argv[1], "help") || strcmp(argv[1], "?") == 0)) {
		print_help_message(argv[0]);
		exit(0);
	}
	
	int c;
	while ((c = getopt(argc, argv, "s:fc:n:a:m:o:")) != EOF) {
		switch(c) {
		case 's':
			if ((f_stats = fopen(optarg, "w")) == NULL) {
				fprintf(stderr, "%s: Could not open statistics file %s " \
						"for writing\n", argv[0], optarg);
				exit(1);
			}
			break;
		case 'a':
			if ((f_align = fopen(optarg, "w")) == NULL) {
				fprintf(stderr, "%s: Could not open align file %s " \
						"for writing\n", argv[0], optarg);
				exit(1);
			}
			break;
		case 'f':
			single_flag = 1;
			break;
		case 'c':
			align_score_cutoff = atoi(optarg);
			if (align_score_cutoff <= 0) {
				fprintf(stderr, "%s: Only positive align score cutoffs " \
						"allowed\n", argv[0]);
				exit(1);
			}
			break;
		case 'n':
			align_top_n = atoi(optarg);
			if (align_top_n < 0) {
				fprintf(stderr, "%s: Only positive n allowed for aligning " \
						"top n candidates\n", argv[0]);
				exit(1);
			}
			break;
		case 'm':
			min_unclustered_count = atoi(optarg);
			if (min_unclustered_count < 0) {
				fprintf(stderr, "%s: Only positive count mins allowed\n",
						argv[0]);
				exit(1);
			}
			break;
		case 'o':
			f_out = fopen(optarg, "w");
			if (!f_out) {
				fprintf(stderr, "Could not open %s for writing\n", optarg);
				exit(1);
			}
			break;
		default:
			fprintf(stderr, "%s: Unrecognized command line option (%c)\n",
					argv[0], c);
			exit(1);
		}
	}
	
	if (optind == argc) {
		f_in = stdin;
	} else if (optind == argc - 1) {
		f_in = fopen(argv[optind], "r");
		if (f_in == NULL) {
			fprintf(stderr, "%s: Could not access file %s\n", argv[0], argv[optind]);
			exit(1);
		}
	} else {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	///PART 1 BEGINS
	fprintf(stderr, "Building data from file...\n");
	
	build_table();
	extract_entries();
	num_entries = hash_entries(h);
	
	///PART 2 BEGINS
	fprintf(stderr, "Finding mismatch counts...\n");
	
	index_link = hash_create();
	lists = calloc(num_entries, sizeof(list*));
	mm1_counts = calloc(num_entries, sizeof(int));
	mm2_counts = calloc(num_entries, sizeof(int));
	totals = calloc(num_entries, sizeof(int));
	mm1_totals = calloc(num_entries, sizeof(int));
	mm2_totals = calloc(num_entries, sizeof(int));
	used = calloc(num_entries, sizeof(char));
	core = calloc(num_entries, sizeof(char));
	
	int i, j, k;
	int ij, ijk;
	for (i = 0; i < num_entries; i++) {
		lists[i] = list_create();
		lists[i]->sorted = 0;
		mm1_counts[i] = overlap(entries[i].key, lists[i]);	//still uses h
		list_sort_by_key(lists[i]);
		mm2_counts[i] = 0;
		totals[i] = entries[i].value;
		mm1_totals[i] = 0;
		mm2_totals[i] = 0;
		used[i] = 0;
		core[i] = 0;
		hash_create_entry(index_link, entries[i].key, i);
	}
	hash_destroy(h);
	
	///PART 3 BEGINS
	fprintf(stderr, "Forming initial clusters...");
	
	int max_mmc, core_total;
	int fraction;
	char *max_seq, *my_seq, *their_seq;
	
	if (single_flag) {
		fprintf(stderr, " (forced single clustering)\n");
		//this is the algorithm where the counts are added completely
		// to the core sequence (and not split up between potential cores)
		for (i = 0; i < num_entries; i++) {
			if (used[i]) {
				//don't try clustering if this entry has already been clustered
				continue;
			}
			
			//find the largest mismatch entry in our group
			max_mmc = mm1_counts[i];
			max_seq = entries[i].key;
			for (j = 0; j < lists[i]->size; j++) {
				ij = hash_lookup(index_link, lists[i]->entries[j]->key);
				if ((c = mm1_counts[ij]) > max_mmc) {
					max_mmc = c;
					max_seq = entries[ij].key;
				}
			}
			
			//now max_seq is the sequence that we cluster around
			c = hash_lookup(index_link, max_seq);
			//c is the index to cluster around, so cluster together lists[c]
			
			if (lists[c]->size == 0) {
				//this turns out to be an empty cluster... skip it
				continue;
			}
			
			for (j = 0; j < lists[c]->size; j++) {
				totals[c] += lists[c]->entries[j]->value;
				mm1_totals[c] += lists[c]->entries[j]->value;
			}
			
			used[c] = 1;	//mark the core as used
			core[c] = 1;	//mark this as the core
			for (j = 0; j < lists[c]->size; j++) {
				my_seq = lists[c]->entries[j]->key;
				ij = hash_lookup(index_link, my_seq);
				
				used[ij] = 1;	//mark this entry as used
				
				//update the other lists so they know
				// our entry is no longer available
				for (k = 0; k < lists[ij]->size; k++) {
					their_seq = lists[ij]->entries[k]->key;
					if (strcmp(their_seq, entries[c].key) == 0) {
						//don't remove from our own cluster list
						continue;
					}
					ijk = hash_lookup(index_link, their_seq);
					list_remove_entry(lists[ijk], my_seq);
					list_remove_entry(lists[ijk], entries[c].key);
				}
			}
		}
	} else {
		fprintf(stderr, "\n");
		//this is the algorithm where counts will be split up between
		// all potential cores
		for (i = 0; i < num_entries; i++) {
			if (core[i]) {
				//this entry has already been considered as a core
				continue;
			}
			if (lists[i]->size == 0) {
				//empty cluster
				continue;
			}
			
			//find the largest mismatch entry in our group
			max_mmc = mm1_counts[i];
			max_seq = entries[i].key;
			for (j = 0; j < lists[i]->size; j++) {
				ij = hash_lookup(index_link, lists[i]->entries[j]->key);
				if ((c = mm1_counts[ij]) > max_mmc && !used[ij]) {
					//a used entry is no longer available
					// to become a core
					max_mmc = c;
					max_seq = entries[ij].key;
				}
			}
			
			//now max_seq is the sequence that we cluster around
			c = hash_lookup(index_link, max_seq);
			//c is the index to cluster around, so cluster together lists[c]
			
			if (used[c]) {
				//cannot make this the core
				continue;
			}
			
			used[c] = 1;	//mark the core as used
			core[c] = 1;	//mark this as the core
			for (j = 0; j < lists[c]->size; j++) {
				my_seq = lists[c]->entries[j]->key;
				ij = hash_lookup(index_link, my_seq);
				
				used[ij] = 1;	//mark this entry as used
			}
		}
		//all the cores have been found, now we have to go through
		// again and assign the totals
		for (i = 0; i < num_entries; i++) {
			if (core[i] || !used[i]) {
				continue;
			}
			
			core_total = 0;
			for (j = 0; j < lists[i]->size; j++) {
				ij = hash_lookup(index_link, lists[i]->entries[j]->key);
				if (core[ij]) {
					core_total += entries[ij].value;
				}
			}
			
			fraction = 0;
			for (j = 0; j < lists[i]->size; j++) {
				ij = hash_lookup(index_link, lists[i]->entries[j]->key);
				if (core[ij]) {
					//would like to be able to round this, but
					// slows down computation significantly
					fraction = entries[i].value * entries[ij].value / core_total;
					
					totals[ij] += fraction;
					mm1_totals[ij] += fraction;
					
					ijk = list_find(lists[ij], entries[i].key);
					lists[ij]->entries[ijk]->value = fraction;
				}
			}
		}
	}
	
	///PART 4 BEGINS
	fprintf(stderr, "Forming secondary clusters...\n");
	//at this point, all of the !used[i] entries correspond
	// to an unclustered entry
	//in_range will keep track of which cores are in the double
	// degeneracy range of the unclustered entry
	/*
	 * To form secondary clusters, create a hash of all first mismatches
	 * of each of the cores, then check if any mismatches of the
	 * entries are in the hash, and assign it to any core it matches
	 */
	
	int in_range_count;
	char *in_range = calloc(num_entries, sizeof(char));
	
	hash* first_mismatch = hash_create();
	int len;
	char oldchr;
	list** first_mismatch_list = calloc(num_entries, sizeof(list*));
	
	//generate core mismatch hash
	for (i = 0; i < num_entries; i++) {
		first_mismatch_list[i] = NULL;
		if (core[i]) {
			len = strlen(entries[i].key);
			hash_create_entry(first_mismatch, entries[i].key, i);
			
			for (j = 0; j < len; j++) {
				oldchr = entries[i].key[j];
				for (k = 0; k < 5; k++) {
					entries[i].key[j] = "ACGTN"[k];
					if (entries[i].key[j] == oldchr) {
						continue;
					}
					
					if ((c = hash_lookup(first_mismatch, entries[i].key)) >= 0) {
						if (first_mismatch_list[c] == NULL) {
							first_mismatch_list[c] = list_create();
							first_mismatch_list[c]->sorted = 0;
						}
						list_create_entry(first_mismatch_list[c],
								entries[i].key, i);
					} else {
						hash_create_entry(first_mismatch, entries[i].key, i);
					}
				}
				entries[i].key[j] = oldchr;
			}
		}
	}
	
	//find mismatches of remaining entries in hash
	for (i = 0; i < num_entries; i++) {
		if (!used[i]) {
			core_total = 0;
			len = strlen(entries[i].key);
			for (j = 0; j < num_entries; j++) {
				in_range[j] = 0;
			}
			
			for (j = 0; j < len; j++) {
				oldchr = entries[i].key[j];
				for (k = 0; k < 5; k++) {
					entries[i].key[j] = "ACGTN"[k];
					if (entries[i].key[j] == oldchr) {
						continue;
					}
					if ((c = hash_lookup(first_mismatch, entries[i].key)) >= 0) {
						if (!in_range[c]) {
							core_total += entries[c].value;
						}
						in_range[c] = 1;
						if (first_mismatch_list[c]) {
							for (ij = 0; ij < first_mismatch_list[c]->size; ij++) {
								//ijk is another core corresponding to the degen
								// but the degen might not be ours... so check
								if (strcmp(first_mismatch_list[c]->entries[ij]->key,
										entries[i].key)) {
									continue;
								}
								ijk = first_mismatch_list[c]->entries[ij]->value;
								if (!in_range[ijk]) {
									core_total += entries[ijk].value;
								}
								in_range[ijk] = 1;
							}
						}
						used[i] = 1;
					}
				}
				entries[i].key[j] = oldchr;
			}
			
			if (!used[i] || core_total == 0) {
				continue;
			}
			
			for (j = 0; j < num_entries; j++) {
				if (in_range[j]) {
					fraction = (entries[i].value * entries[j].value) / core_total;
					
					totals[j] += fraction;
					mm2_totals[j] += fraction;
					mm2_counts[j]++;
					list_create_entry(lists[j], entries[i].key, fraction);
				}
			}
		}
	}
	
	for (i = 0; i < num_entries; i++) {
		if (first_mismatch_list[i]) {
			list_destroy(first_mismatch_list[i]);
		}
	}
	free(first_mismatch_list);
	hash_destroy(first_mismatch);
	
	///OUTPUT CLUSTERS
	//sort entries by total count
	entry* sorted_entries = calloc(num_entries, sizeof(entry));
	for (i = 0; i < num_entries; i++) {
		sorted_entries[i].key = entries[i].key;
		sorted_entries[i].value = totals[i];
	}
	qsort(sorted_entries, num_entries, sizeof(entry), entry_comparator);
	
	if (f_out) {
		fprintf(f_out, "#core sequence\tscore\ttotal count\tcore count" \
				"\tmm1 count\tmm2 count\t[sequence:count:mmcount]...\n");
		for (i = 0; i < num_entries; i++) {
			ij = hash_lookup(index_link, sorted_entries[i].key);
			if (core[ij]) {
				fprintf(f_out, "%s\t%d:%d\t%d\t%d\t%d\t%d\t", entries[ij].key,
						mm1_counts[ij], mm2_counts[ij], totals[ij],
						entries[ij].value, mm1_totals[ij], mm2_totals[ij]);
				for (j = 0; j < lists[ij]->size; j++) {
					ijk = hash_lookup(index_link, lists[ij]->entries[j]->key);
					fprintf(f_out, "%s:%d:%d", lists[ij]->entries[j]->key,
							lists[ij]->entries[j]->value, mm1_counts[ijk] +
							mm2_counts[ijk]);
					if (j != lists[ij]->size - 1) {
						fprintf(f_out, ",");
					}
				}
				fprintf(f_out, "\n");
			}
		}
		
		for (i = 0; i < num_entries; i++) {
			if (!used[i]) {
				if (totals[i] > min_unclustered_count) {
					fprintf(f_out, "%s\t%d:%d\t%d\t%d\t%d\t%d\n", entries[i].key,
							mm1_counts[i], mm2_counts[i], totals[i],
							entries[i].value, mm1_totals[i], mm2_totals[i]);
				} else {
					break;
				}
			}
		}
	}
	
	printf("#core sequence\tscore\ttotal count\tcore count" \
			"\tmm1 count\tmm2 count\t[sequence:count:mmcount]...\n");
	for (i = 0; i < num_entries; i++) {
		ij = hash_lookup(index_link, sorted_entries[i].key);
		if (core[ij]) {
			printf("%s\t%d:%d\t%d\t%d\t%d\t%d\t", entries[ij].key,
					mm1_counts[ij], mm2_counts[ij], totals[ij],
					entries[ij].value, mm1_totals[ij], mm2_totals[ij]);
			for (j = 0; j < lists[ij]->size; j++) {
				ijk = hash_lookup(index_link, lists[ij]->entries[j]->key);
				printf("%s:%d:%d", lists[ij]->entries[j]->key,
						lists[ij]->entries[j]->value, mm1_counts[ijk] +
						mm2_counts[ijk]);
				if (j != lists[ij]->size - 1) {
					printf(",");
				}
			}
			printf("\n");
		}
	}
	
	for (i = 0; i < num_entries; i++) {
		if (!used[i]) {
			if (totals[i] > min_unclustered_count) {
				printf("%s\t%d:%d\t%d\t%d\t%d\t%d\n", entries[i].key,
						mm1_counts[i], mm2_counts[i], totals[i],
						entries[i].value, mm1_totals[i], mm2_totals[i]);
			} else {
				break;
			}
		}
	}
	
	if (f_stats == NULL) {
		return 0;	//if no stats are required, just exit here
	}
	
	///PART 5 BEGINS
	//align generally without expected barcodes
	align_general();
	
	///STATISTICS TAKING
	//a lot of extra variables will need to be created to keep track
	// of the various clustering possibilities of each entry
	//proximal clusters keep track of the number of clusters that are
	// close to ours
	//largest, smallest cluster sizes are kept track of
	//all cluster sizes are also recorded
	int *proximal_clusters = calloc(num_entries, sizeof(int));
	int *largest_cluster_size = calloc(num_entries, sizeof(int));
	int *smallest_cluster_size = calloc(num_entries, sizeof(int));
	int **all_cluster_sizes = calloc(num_entries, sizeof(int*));
	int largest, smallest;
	
	fprintf(stderr, "Generating cluster statistics...\n");
	k = (num_entries / 10);	//for recording progress in stderr
	for (i = 0; i < num_entries; i++) {
		proximal_clusters[i] = 0;
		if (!core[i]) {
			if (used[i]) {
				ijk = 1;
			} else {
				ijk = 2;
			}
			in_range_count = 0;
			largest = 0;
			smallest = 999999;
			for (j = 0; j < num_entries; j++) {
				if (core[j] && in_n_degen(entries[i].key, entries[j].key, ijk)) {
					if (totals[j] > largest) {
						largest = totals[j];
					}
					if (totals[j] < smallest) {
						smallest = totals[j];
					}
					in_range_count++;
					in_range[j] = 1;
				} else {
					in_range[j] = 0;
				}
			}
			all_cluster_sizes[i] = calloc(in_range_count, sizeof(int));
			ij = 0;
			for (j = 0; j < num_entries; j++) {
				if (in_range[j]) {
					all_cluster_sizes[i][ij++] = totals[j];
				}
			}
			proximal_clusters[i] = in_range_count;
			largest_cluster_size[i] = largest;
			if (smallest < 999999) {
				smallest_cluster_size[i] = smallest;
			} else {
				smallest_cluster_size[i] = 0;
			}
		}
		if (i % k == 0) {
			fprintf(stderr, "\t%d of %d performed [%d%%]\n", i,
					num_entries, (i*10/k));	//log how much we finished
		}
	}
	
	fprintf(f_stats, "#Degeneracy\tSequence Count\tSequence\tProximal Clusters\t" \
			"Largest Proximal Cluster\tSmallest Proximal Cluster\t" \
			"All proximal cluster sizes\n");
	
	for (i = 0; i < num_entries; i++) {
		if (core[i] || proximal_clusters[i] == 0) {
			continue;	//cores and uncluster-able entries are excluded
		}
		fprintf(f_stats, "%d\t%d\t%s\t%d\t%d\t%d\t", 2 - used[i],
				entries[i].value, entries[i].key, proximal_clusters[i],
				largest_cluster_size[i], smallest_cluster_size[i]);
		for (j = 0; j < proximal_clusters[i]; j++) {
			fprintf(f_stats, "%d", all_cluster_sizes[i][j]);
			if (j != proximal_clusters[i] - 1) {
				fprintf(f_stats, ",");
			}
		}
		fprintf(f_stats, "\n");
	}
	
	return 0;
};
