#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "include.h"
#include "hash.h"

static int trim_start = 0;
static int trim_end = 0;
static int trunc = 0;
static int fastq_flag = 0;
static int qsec_flag = 0;
static int clean_flag = 0;
static int quality_lo = 0;
static int quality_hi = 40;
static int quality_adj = 64;
static int quality_cutoff = 10;
static FILE *f_in = NULL;
static FILE *f_out = NULL;

static void print_usage_message(char *argv0) {
	fprintf(stderr, "usage: %s [-i 5':3'] [-u n] [-f|q] [-o output_file] " \
			"[-m S|X|I|J|L] [-c] [input_file]\n", argv0);
}

static void print_help_message(char *argv0) {
	printf("usage: %s [-i 5':3'] [-u n] [-f|q] [-o output_file] " \
			"[-m S|X|I|J] [-c] [input_file]\n", argv0);
	printf("input_file specifies the file with entries either in " \
			"qseq or fastq format; if none provided, reads stdin\n\n");
	printf("  -i X:Y\t trim X bases on 5' end and Y " \
			"bases on 3' end\n");
	printf("  -u N\t\t truncate sequence to N bases\n");
	printf("  -f\t\t force input file to be parsed as a fastq file\n");
	printf("  -q\t\t force input file to be parsed as a qsec file\n");
	printf("  -o FILE\t additionally prints results to FILE\n");
	printf("  -m S|X|I|J\t filters by Sanger, Solexa, " \
			"Illumina 1.3+, Illumina 1.5+ (respectively)\n");
	printf("  -c\t\t only unfiltered results will be reported " \
			"in output\n");
}

/**
 * Prints the results of the count to the file specified in the
 * global variable "f_out", and then to stdout
 * The f_out is set in the count_main
 */
static void print_results(int total_reads, int unique_reads, int filtered, entry* entries) {
	if (f_out != NULL) {
		fprintf(f_out, "#trim: %d:%d\n", trim_start, trim_end);
		fprintf(f_out, "#truncate: %d\n", trunc);
		fprintf(f_out, "#quality lo: %d\thi: %d\tadj: %d\t" \
				"cutoff: %d\n", quality_lo, quality_hi, quality_adj, quality_cutoff);
		fprintf(f_out, "#total reads: %d\n", total_reads);
		fprintf(f_out, "#unique reads: %d\n", unique_reads);
		fprintf(f_out, "#filtered bases: %d\n", filtered);
		
		fprintf(f_out, "#read\tcount\n");
		//assuming entries is as big as unique reads
		int i;
		for (i = 0; i < unique_reads; i++) {
			fprintf(f_out, "%s\t%d\n", entries[i].key, entries[i].value);
		}
	}
	f_out = stdout;
	fprintf(f_out, "#trim: %d:%d\n", trim_start, trim_end);
	fprintf(f_out, "#truncate: %d\n", trunc);
	fprintf(f_out, "#quality lo: %d\thi: %d\tadj: %d\t" \
			"cutoff: %d\n", quality_lo, quality_hi, quality_adj, quality_cutoff);
	fprintf(f_out, "#total reads: %d\n", total_reads);
	fprintf(f_out, "#unique reads: %d\n", unique_reads);
	fprintf(f_out, "#filtered bases: %d\n", filtered);
	
	fprintf(f_out, "#read\tcount\n");
	//assuming entries is as big as unique reads
	int i;
	for (i = 0; i < unique_reads; i++) {
		fprintf(f_out, "%s\t%d\n", entries[i].key, entries[i].value);
	}
}

/**
 * Filters a sequence given the corresponding quality string
 * Quality filtered is based on the global quality_ variables
 * If the phred score of a base does not meet the cutoff, it is
 * changed to a 'N'
 * The number of these such changes are recorded and returned
 */
static int quality_filter(char *seq, char *qual) {
	if (strlen(seq) != strlen(qual)) {
		fprintf(stderr, "Sequence length and quality length mismatch\n");
		exit(3);
	}
	
	int replaces = 0;
	char c;
	int score;
	
	while ((c = *seq)) {
		score = *qual - quality_adj;
		if (score < quality_cutoff) {
			*seq = 'N';
			replaces++;
		}
		if (*seq == '.') {
			*seq = 'N';
		}
		seq++;
		qual++;
	}
	
	return replaces;
}

/**
 * Processes a fastq formatted file
 * Pulls the DNA sequence and quality string from the file
 * Trims and truncates both the above strings
 * Keeps track of the number of appearances of the DNA sequence
 * in a hash table
 * Prints the results using the above print method
 */
static void process_fastq() {
	hash *h = hash_create();
	
	char buffer[1024];
	char *p;
	char seq[1024];
	char *qual;
	int s;
	int filt_i;
	int linetype = 0;
	
	int records = 0;
	int filtered = 0;
	
	while (fgets(buffer, 1024, f_in)) {
		s = strlen(buffer);
		if (buffer[s-1] == '\n') {
			buffer[s-1] = '\0';
		}
		
		if (linetype == 1) {
			//these are DNA sequences
			//trim the entry
			p = &buffer[trim_start];
			if (strlen(buffer) - trim_end < 0) {
				fprintf(stderr, "Trimmed beyond sequence data\n");
				exit(3);
			}
			p[strlen(p) - trim_end] = '\0';
			
			//truncate the entry
			if (trunc > 0) {
				p[trunc] = '\0';
			}
			
			strcpy(seq, p);
			seq[strlen(p)] = '\0';
		} else if (linetype == 3) {
			//These are corresponding quality strings
			//trim the entry
			qual = &buffer[trim_start];
			if (strlen(buffer) - trim_end < 0) {
				fprintf(stderr, "Trimmed beyond sequence data\n");
				exit(3);
			}
			qual[strlen(qual) - trim_end] = '\0';
			
			//truncate the entry
			if (trunc > 0) {
				qual[trunc] = '\0';
			}
			
			filt_i = quality_filter(seq, qual);
			
			filtered += filt_i;
			
			if (!clean_flag || filt_i == 0) {
				hash_increment(h, seq);
			}
			records++;
		}
		
		linetype = (linetype+1) % 4;
	}
	
	int unique = hash_entries(h);
	entry* sorted = hash_sorted_entries(h);
	
	print_results(records, unique, filtered, sorted);
	
	hash_destroy(h);
}

/**
 * Processes a qsec formatted file
 * Pulls the DNA sequence and quality string from the file
 * Trims and truncates both the above strings
 * Keeps track of the number of appearances of the DNA sequence
 * in a hash table
 * Prints the results using the above print method
 */
static void process_qsec() {
	hash *h = hash_create();
	
	char buffer[1024];
	char *seq;
	char *qual;
	int s;
	int filt_i;
	
	int sequence_col = 8;
	
	int records = 0;
	int filtered = 0;
	
	while (fgets(buffer, 1024, f_in)) {
		s = strlen(buffer);
		if (buffer[s-1] == '\n') {
			buffer[s-1] = '\0';
		}
		int i;
		seq = buffer;
		for (i = 0; i < sequence_col; i++) {
			seq = strchr(seq, '\t') + 1;
		}
		qual = strchr(seq, '\t') + 1;
		
		strchr(seq, '\t')[0] = '\0';
		strchr(qual, '\t')[0] = '\0';
		
		//trim the entry
		seq = seq + trim_start;
		if (strlen(seq) - trim_end < 0) {
			fprintf(stderr, "Trimmed beyond sequence data\n");
			exit(3);
		}
		seq[strlen(seq) - trim_end] = '\0';
		
		qual = qual + trim_start;
		if (strlen(qual) - trim_end < 0) {
			fprintf(stderr, "Trimmed beyond sequence data\n");
			exit(3);
		}
		qual[strlen(qual) - trim_end] = '\0';
		
		//truncate the entry
		if (trunc > 0) {
			if (trunc < strlen(seq)) {
				seq[trunc] = '\0';
			}
			if (trunc < strlen(qual)) {
				qual[trunc] = '\0';
			}
		}
		
		filt_i = quality_filter(seq, qual);
		
		filtered += filt_i;
		
		if (!clean_flag || filt_i == 0) {
			hash_increment(h, seq);
		}
		records++;
	}
	
	int unique = hash_entries(h);
	entry* sorted = hash_sorted_entries(h);
	
	print_results(records, unique, filtered, sorted);
	
	hash_destroy(h);
}

/**
 * Entry point of the count method
 * Most of work here is to just parse the command line options and
 * set the global variables as specified by the user
 * Once that is over, it simply relays the information to the
 * process_fastq or process_qsec, depending on which is appropriate, to
 * do the actual counting work
 */
int count_main(int argc, char **argv) {
	if (argc > 1 && (strcmp(argv[1], "?") == 0 || strstr(argv[1], "help"))) {
		print_help_message(argv[0]);
		exit(0);
	}
	
	int c;
	char *opt;
	
	while ((c = getopt(argc, argv, "i:u:fqo:m:c")) != EOF) {
		switch(c) {
		case 'i':
			//trim request
			opt = strchr(optarg, ':');
			if (opt == NULL) {
				print_usage_message(argv[0]);
				exit(1);
			}
			opt[0] = '\0';
			trim_start = atoi(optarg);
			trim_end = atoi((opt+1));
			break;
		case 'u':
			//truncate request
			trunc = atoi(optarg);
			break;
		case 'f':
			fastq_flag = 1;
			break;
		case 'q':
			qsec_flag = 1;
			break;
		case 'o':
			f_out = fopen(optarg, "w");
			if (!f_out) {
				fprintf(stderr, "Could not access %s for writing\n", optarg);
				exit(1);
			}
			break;
		case 'm':
			switch(optarg[0]) {
			case 'S':
				quality_lo = 0;
				quality_hi = 40;
				quality_adj = 33;
				break;
			case 'X':
				quality_lo = -5;
				quality_hi = 40;
				quality_adj = 64;
				break;
			case 'I':
				quality_lo = 0;
				quality_hi = 40;
				quality_adj = 64;
				break;
			case 'J':
				quality_lo = 3;
				quality_hi = 40;
				quality_adj = 64;
				break;
			case 'L':
				quality_lo = 0;
				quality_hi = 41;
				quality_adj = 33;
				break;
			default:
				fprintf(stderr, "Unrecognized quality filter method (%c)\n", optarg[0]);
				exit(1);
			}
			if (optarg[1] != '\0') {
				quality_cutoff = atoi(&optarg[1]);
				if (quality_cutoff <= 0) {
					fprintf(stderr, "Invalid quality cutoff: %d\n", quality_cutoff);
					exit(1);
				}
			}
			break;
		case 'c':
			clean_flag = 1;
			break;
		default:
			print_usage_message(argv[0]);
			exit(1);
		}
	}
	
	if (optind == argc) {
		f_in = stdin;
	} else if (optind == argc - 1) {
		f_in = fopen(argv[optind], "r");
		if (!f_in) {
			fprintf(stderr, "Could not read from file %s\n", argv[optind]);
		}
	} else {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	if (fastq_flag && qsec_flag) {
		//Cannot force processing as both, must be one or the other
		fprintf(stderr, "%s: Both fastq and qsec flags indicated\n", argv[0]);
		exit(1);
	}
	
	if (!(fastq_flag || qsec_flag)) {
		if (f_in == stdin) {
			fprintf(stderr, "Reading from stdin requires -f or -q\n");
			exit(2);
		}
		
		if (strrchr(argv[optind], '.') == NULL) {
			fprintf(stderr, "Cannot determine file format of %s\n", argv[optind]);
			exit(2);
		}
		if (strcmp(strrchr(argv[optind], '.'), ".fastq") == 0 ||
			strcmp(strrchr(argv[optind], '.'), ".fq") == 0) {
			fastq_flag = 1;
		} else if (strcmp(strrchr(argv[optind], '.'), ".txt") == 0) {
			qsec_flag = 1;
		} else {
			fprintf(stderr, "Expected .fastq/.fq or .txt file (%s)\n", argv[optind]);
			exit(2);
		}
	}
	
	if (fastq_flag) {
		process_fastq();
	} else if (qsec_flag) {
		process_qsec();
	} else {
		fprintf(stderr, "Could not determine input type (fastq or qseq)\n");
		exit(2);
	}
	
	return 0;
};
