#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "hash.h"

/**
 * The hashing algorithm used is taken from public domain (sdbm)
 * It may be desirable to switch to another hash function in particular
 * barseq runs, but this is good for most cases
 */
long hash_slot(char* str) {
	//sdbm hashing algorithm
	long long h = 0;
	int c;
	
	while ((c = *str++)) {
		h = c + (h << 6) + (h << 16) - h;
		h = h % HASH_SLOTS;
	}
	
	return h;
};

int hash_slot_nmer(char *word, int n) {
	int index=0;
	int base;
	int i, j=0;
	while (*word && (j++ < n)) {
		for (i = 0; i < 5; i++) {
			if (*word == "ACGTN"[i]) {
				base = i;
				break;
			}
		}
		//i = 5i + next
		index = (index << 2) + index + base;
		
		word++;
	}
	
	return index;
};

/**
 * Creates an empty hash table for use with each of the other
 * hash functions
 * Requires no parameters
 */
hash* hash_create() {
	hash* ret = malloc(sizeof(hash));
	ret->slots = calloc(HASH_SLOTS, sizeof(list*));
	
	if (ret->slots == NULL) {
		fprintf(stderr, "Could not allocate enough space\n");
	}
	
	long i;
	for (i = 0; i < HASH_SLOTS; i++) {
		ret->slots[i] = list_create();
	}
	
	return ret;
};

/**
 * Frees all memory being used by a supplied hash table
 */
void hash_destroy(hash* h) {
	long i;
	for (i = 0; i < HASH_SLOTS; i++) {
		if (h->slots[i]) {
			list_destroy(h->slots[i]);
		}
	}
	free(h->slots);
	free(h);
};

/**
 * Performs the hash lookup of the supplied key in the hash table
 */
int hash_lookup(hash* h, char *key) {
	long i = hash_slot(key);
	int j = list_find(h->slots[i], key);
	
	if (j < 0) {
		//not found in the hash table
		return -1;
	}
	
	return h->slots[i]->entries[j]->value;
};

/**
 * Increments the value of the hash entry associated with the key
 * Note: If a value for the key does not exist, an entry is created
 * for the key and given an initial value of "1"
 */
void hash_increment(hash *h, char *key) {
	long i = hash_slot(key);
	
	list_increment(h->slots[i], key);
};

/**
 * Decrements the value of the hash entry associated with the key
 * Note: If there is no value associated with the key, nothing happens
 */
void hash_decrement(hash *h, char *key) {
	long i = hash_slot(key);
	
	list_decrement(h->slots[i], key);
};

/**
 * Sets the value of the provided key in the hash
 * Creates an entry for key if one does not exist already
 */
void hash_set_value(hash *h, char *key, int value) {
	long i = hash_slot(key);
	
	list_set_value(h->slots[i], key, value);
};

/**
 * Returns the number of unique keys in this hash table
 */
int hash_entries(hash* h) {
	int ret = 0;
	long i;
	for (i = 0; i < HASH_SLOTS; i++) {
		ret += h->slots[i]->size;
	}
	
	return ret;
};

/**
 * Creates a new (key, value) pair in the hash table
 * Note: Does not check if the pair exists already; if it does, the
 * entry is added anyway (as a duplicate) and strange behaviour
 * may occur
 * The caller is given responsibility of checking existence in the
 * hash table; this reduces overhead significantly
 */
void hash_create_entry(hash *h, char *key, int value) {
	long i = hash_slot(key);
	
	list_create_entry(h->slots[i], key, value);
};

/**
 * Removes the entry associated with the supplied key in the hash table
 * If there is no entry for the supplied key, nothing happens
 */
void hash_remove_entry(hash *h, char *key) {
	long i = hash_slot(key);
	
	list_remove_entry(h->slots[i], key);
};
