#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "list.h"

/* Default list size; kept low for low memory consumption at the cost
   of performance - May be increased for server computing */
#define	LIST_INIT_SIZE	1

/**
 * Creates a new list for use with the below list methods
 * Lists are dynamic arrays of (key, value) pairs
 * For improved performance, the entries are sorted by the 'key' field
 * (by alphabetical order) to improve find times
 * By setting the list's sorted field, it is possible to stop the
 * sorted behaviour requirement
 */
list* list_create() {
	list* ret = malloc(sizeof(list));
	
	ret->entries = calloc(LIST_INIT_SIZE, sizeof(entry*));
	ret->size = 0;
	ret->max_size = LIST_INIT_SIZE;
	ret->sorted = 1;
	
	int i;
	for (i = 0; i < ret->max_size; i++) {
		ret->entries[i] = NULL;
	}
	
	return ret;
}

/**
 * Frees all memory associated with the supplied list
 */
void list_destroy(list* l) {
	int i;
	for (i = 0; i < l->size; i++) {
		if (l->entries[i]) {
			if (l->entries[i]->key) {
				free(l->entries[i]->key);
			}
			free(l->entries[i]);
		}
	}
	free(l->entries);
	free(l);
}

/**
 * For use with binary search comparator to see if
 * the given key matches the given entry
 */
static int bsearch_comparator(const void *a1, const void *a2) {
	char *key = (char*) a1;
	entry **e = (entry**) a2;
	
	return strcmp(key, (*e)->key);
}

/**
 * Entry point for finding a key in a list
 * Goes through the list and uses ordered property to find if item
 * is in list
 * If list is unsorted, it must check the whole list
 */
int list_find(list* l, char* key) {
	//returns the list's index for the supplied key index
	//returns -1 if the key is not in the list
	if (l->size == 0) {
		return -1;
	}
	
	if (l->sorted) {
		entry** e = (entry**) bsearch(key, l->entries, l->size,
				sizeof(entry*), bsearch_comparator);
		if (e) {
			return (e - l->entries);
		} else {
			return -1;
		}
	}
	
	int i;
	for (i = 0; i < l->size; i++) {
		if (strcmp(l->entries[i]->key, key) == 0) {
			return i;
		}
	}
	
	return -1;
}

/**
 * Increments the value associated with key in this list
 * If there is no value already present, a new value is created and
 * inserted with an initial value of "1"
 */
void list_increment(list* l, char* key) {
	int i = list_find(l, key);
	
	if (i >= 0) {
		l->entries[i]->value++;
	} else {
		list_create_entry(l, key, 1);
	}
}

/**
 * Increments the value associated with key by n in this list
 * If no value already exists, it is created with value "value"
 */
void list_increment_n(list* l, char* key, int value) {
	int i = list_find(l, key);
	
	if (i >= 0) {
		l->entries[i]->value += value;
	} else {
		list_create_entry(l, key, value);
	}
}

/**
 * Decrements the value associated with key in this list
 * If there is no value associated with key, nothing is done
 */
void list_decrement(list* l ,char* key) {
	int i = list_find(l, key);
	
	if (i >= 0) {
		l->entries[i]->value--;
	}
}

/**
 * Sets the value of the supplied key to value
 * If no such entry exists, the entry is created
 */
void list_set_value(list* l, char* key, int value) {
	int i = list_find(l, key);
	
	if (i >= 0) {
		l->entries[i]->value = value;
	} else {
		list_create_entry(l, key, value);
	}
}

/**
 * Helps find where a new entry should go in a list
 * The returned value is the index where it is found to go
 * This preserves the ordering/sorted property of the list
 */
static int list_insert_location(list *l, char* key) {
	//finds the index in list where an entry would go
	if (l->size == 0) {
		return 0;
	}
	
	int i;
	int len = l->size;
	for (i = 0; i < len; i++) {
		if (strcmp(l->entries[i]->key, key) >= 0) {
			break;
		}
	}
	return i;
}

/**
 * Creates a new (key, value) pair in this list
 * Note: Does not check if the pair exists already; if it does, the
 * entry is added anyway (as a duplicate) and strange behaviour
 * may occur
 * The caller is given responsibility of checking existence in the
 * list; this reduces overhead significantly
 */
void list_create_entry(list *l, char* key, int value) {
	while (l->size >= l->max_size) {
		l->max_size = l->max_size * 2;
		entry **oldentries = l->entries;
		if ((l->entries = calloc(l->max_size, sizeof(entry*))) == NULL) {
			fprintf(stderr, "Ran out of memory\n");
			exit(2);
		}
		memcpy(l->entries, oldentries, l->size*sizeof(entry*));
		int i;
		for (i = l->size; i < l->max_size; i++) {
			l->entries[i] = NULL;
		}
		free(oldentries);
	}
	
	int new_location;
	if (l->sorted) {
		new_location = list_insert_location(l, key);
		//copy over old stuff
		int i;
		for (i = l->size; i > new_location; i--) {
			l->entries[i] = l->entries[i-1];
		}
	} else {
		new_location = l->size;
	}
	
	int len = strlen(key);
	char *tmp = malloc((len+1) * sizeof(char));
	strcpy(tmp, key);
	tmp[len] = '\0';
	
	entry *newentry = malloc(sizeof(entry));
	newentry->key = tmp;
	newentry->value = value;
	l->entries[new_location] = newentry;
	
	l->size++;
};

/**
 * Removes the entry associated with the supplied key in the list
 * If there is no entry for the supplied key, nothing happens
 */
int list_remove_entry(list *l, char *key) {
	//returns 1 if something was actually removed
	// and 0 if it wasn't in to begin with
	int location = list_find(l, key);
	if (location < 0) {
		return 0;
	}
	entry *e = l->entries[location];
	if (e) {
		if (e->key) {
			free(e->key);
		}
		free(e);
	}
	
	int i;
	for (i = location; i < l->size - 1; i++) {
		l->entries[i] = l->entries[i+1];
	}
	l->entries[l->size] = NULL;
	
	l->size--;
	return 1;
};
