#include <stdio.h>
#include <unistd.h>
#include "include.h"
#include "hash.h"
#include "list.h"
#include "alignment.h"

static list *barcode_sequences = NULL;
static char **barcode_details = NULL;
static list *cluster_sequences = NULL;
static int *cluster_totals = NULL;
static int *cluster_mm1n = NULL;
static int *cluster_mm2n = NULL;
static int *cluster_total_mm0 = NULL;
static int *cluster_total_mm1 = NULL;
static int *cluster_total_mm2 = NULL;
static hash *words = NULL;
static int num_barcode_sequences = 0;
static int num_clusters = 0;
static int **match = NULL;
static char *match_type = NULL;
static char **match_alignment = NULL;
static int align_top_n = 0;
static int align_count_minimum = 0;
static char skip_alignment = 0;

static FILE *f_cluster = NULL;
static FILE *f_barcode = NULL;
static FILE *f_out = NULL;
static int MATCH_NONE = 0;
static int MATCH_CONTAIN = 1;
static int MATCH_ALIGN = 2;
static int MATCH_ALIGNMULTIPLE = 3;
static int MATCH_MM1 = 4;
static int MATCH_MM2 = 5;

static void print_usage_message(char *argv0) {
	fprintf(stderr, "Usage: %s [-c min_count_cutoff] [-n align_top_n] " \
			"[-o output_file] [-s] [cluster_file] barcode_file\n", argv0);
}

static void print_help_message(char *argv0) {
	printf("Usage: %s [-c min_count_cutoff] [-n align_top_n] " \
			"[-o output_file] [-s] [cluster_file] barcode_file\n", argv0);
	printf("Generates a list of best-matching barcodes for " \
			"each cluster sequence\n");
	printf("cluster_file specifies clusters to match to barcodes, if " \
			"unspecified, will read stdin\n");
	printf("barcode_file specifies barcodes available for matching\n\n");
	printf("  -c CUTOFF\t clusters require a count of CUTOFF or more " \
			"to be aligned\n");
	printf("  -n N\t\t clusters try only to align to the " \
			"top N candidate barcodes\n");
	printf("  -o FILE\t additionally outputs results to FILE\n");
	printf("  -s \t\t skips aligning clusters\n");
}

/**
 * Reads the barcodes from the barcodes file into the barcode list,
 * along with the index of where that barcode occurs in the
 * barcode_details variable
 */
static void build_barcode_sequences() {
	//assumes f_barcode is valid and barcode_sequences list is ready
	char buff[4096];
	char *ptr, *ptr2;
	while (fgets(buff, 4095, f_barcode)) {
		if (buff[0] == '#') {
			continue;	//commented line; skip it
		}
		
		ptr = strchr(buff, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (details\tsequence)\n");
			fprintf(stderr, "\t%s\n", buff);
			continue;
		}
		
		*ptr = '\0';
		ptr++;
		
		ptr2 = strchr(ptr, '\n');
		if (ptr2) {
			*ptr2 = '\0';
		}
		
		list_set_value(barcode_sequences, ptr, num_barcode_sequences);
		num_barcode_sequences++;
	}
	
	barcode_details = calloc(num_barcode_sequences, sizeof(char*));
	rewind(f_barcode);
	
	char *tmp;
	int tmp_len, i=0;
	while (fgets(buff, 4095, f_barcode)) {
		if (buff[0] == '#') {
			continue;	//commented line; skip it
		}
		
		ptr = strchr(buff, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (details\tsequence)\n");
			fprintf(stderr, "\t%s\n", buff);
			continue;
		}
		*ptr = '\0';
		
		tmp_len = strlen(buff);
		tmp = malloc(tmp_len * (sizeof(char)+1));
		strcpy(tmp, buff);
		tmp[tmp_len] = '\0';
		
		barcode_details[i] = tmp;
		i++;
	}
	
	if (num_barcode_sequences != barcode_sequences->size) {
		fprintf(stderr, "Warning: there are non-unique " \
				"barcode_sequences in the barcode file\n");
		num_barcode_sequences = barcode_sequences->size;
	}
}

/**
 * Creates all the links from 18mer words to barcodes and places them
 * in the words hash for checking containment
 */
static void build_words() {
	//assumes barcode_sequences list is populated
	int collisions = 0;
	int nmer = 18;	//length of the words
	int len;
	char oldchr;
	int i,j;
	for (i = 0; i < num_barcode_sequences; i++) {
		len = strlen(barcode_sequences->entries[i]->key) - nmer + 1;
		for (j = 0; j < len; j++) {
			oldchr = barcode_sequences->entries[i]->key[j+nmer];
			barcode_sequences->entries[i]->key[j+nmer] = '\0';
			if (hash_lookup(words, &barcode_sequences->entries[i]->key[j]) >= 0) {
				collisions++;
			}
			hash_set_value(words, &barcode_sequences->entries[i]->key[j], i);
			barcode_sequences->entries[i]->key[j+nmer] = oldchr;
		}
	}
	if (collisions) {
		fprintf(stderr, "\tWarning: barcode sequence file has " \
				"%dmer word overlaps\n", nmer);
	}
}

/**
 * Modular method usable in building the clusters to check the
 * format of a cluster entry as it goes along
 */
static void cluster_format_check(char *ptr) {
	if (ptr == NULL) {
		fprintf(stderr, "Error: a cluster entry is not " \
				"in the valid format:\t\n" \
				"seq\tmm1n:mm2n\tmm0\tmm1\tmm2\n");
		exit(1);
	}
}

/**
 * Breaks up the cluster file into the information required to map
 * the clusters
 * The cluster_sequences list is given indices into each of the other
 * arrays which includes more information about each cluster
 */
static void build_clusters() {
	//assumes f_cluster is valid and barcode_sequences list is ready
	//need to store remaining cluster details since stdin cannot be rewinded
	char buff[16384];	buff[16383] = '\0';
	list *lines = list_create();
	lines->sorted = 0;
	char *ptr, *ptr2;
	while (fgets(buff, 16383, f_cluster)) {
		if (buff[0] == '#') {
			continue;	//commented line; skip it
		}
		list_set_value(lines, buff, num_clusters);
		
		ptr = strchr(buff, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (sequence\t(cluster details)...)\n");
			continue;
		}
		*ptr = '\0';
		ptr++;
		
		ptr2 = strchr(buff, '\n');
		if (ptr2) {
			*ptr2 = '\0';
		}
		
		list_set_value(cluster_sequences, buff, num_clusters);
		num_clusters++;
	}
	
	cluster_mm1n = calloc(num_clusters, sizeof(int));
	cluster_mm2n = calloc(num_clusters, sizeof(int));
	cluster_total_mm0 = calloc(num_clusters, sizeof(int));
	cluster_total_mm1 = calloc(num_clusters, sizeof(int));
	cluster_total_mm2 = calloc(num_clusters, sizeof(int));
	cluster_totals = calloc(num_clusters, sizeof(int));
	
	int i, j;
	char *line, *score_ptr, *totals_ptr, *totalmm0_ptr,
			*totalmm1_ptr, *totalmm2_ptr;
	for (j = 0; j < num_clusters; j++) {
		line = lines->entries[j]->key;
		i = lines->entries[j]->value;
		
		score_ptr = strchr(line, '\t');
		cluster_format_check(score_ptr);
		*score_ptr = '\0';
		score_ptr++;
		
		ptr = strchr(score_ptr, ':');
		cluster_format_check(ptr);
		*ptr = '\0';
		ptr++;
		
		totals_ptr = strchr(ptr, '\t');
		cluster_format_check(totals_ptr);
		*totals_ptr = '\0';
		totals_ptr++;
		
		totalmm0_ptr = strchr(totals_ptr, '\t');
		cluster_format_check(totalmm0_ptr);
		*totalmm0_ptr = '\0';
		totalmm0_ptr++;
		
		totalmm1_ptr = strchr(totalmm0_ptr, '\t');
		cluster_format_check(totalmm1_ptr);
		*totalmm1_ptr = '\0';
		totalmm1_ptr++;
		
		totalmm2_ptr = strchr(totalmm1_ptr, '\t');
		cluster_format_check(totalmm2_ptr);
		*totalmm2_ptr = '\0';
		totalmm2_ptr++;
		
		//the cluster may or may not have a list of sequences
		// after it now
		ptr2 = strchr(totalmm2_ptr, '\t');
		if (ptr2) {
			*ptr2 = '\0';
		}
		ptr2 = strchr(totalmm2_ptr, '\n');
		if (ptr2) {
			*ptr2 = '\0';
		}
		
		cluster_mm1n[i] = atoi(score_ptr);
		cluster_mm2n[i] = atoi(ptr);
		cluster_total_mm0[i] = atoi(totalmm0_ptr);
		cluster_total_mm1[i] = atoi(totalmm1_ptr);
		cluster_total_mm2[i] = atoi(totalmm2_ptr);
		cluster_totals[i] = atoi(totals_ptr);
	}
	
	if (num_clusters != cluster_sequences->size) {
		fprintf(stderr, "Warning: there are non-unique " \
				"cluster cores in the cluster file\n");
		num_clusters = cluster_sequences->size;
	}
	
	list_destroy(lines);
}

/**
 * Aligns remaining cluster sequences to barcode sequences with parameters
 * indicated by global variables
 * align_top_n will speed up aligning by only trying to align against
 * the top n candidates
 * align_count_minimum will further speed up aligning by ignoring clusters
 * with small counts
 * If a cluster aligns to multiple barcode sequences, a note of it
 * is made in the match type, and the match_alignment variable is
 * set to indicate which alignments were made
 */
static void align_against_barcode_sequences() {
	fprintf(stderr, "Aligning remaining sequences");
	if (align_top_n > 0) {
		fprintf (stderr, " (top %d candidates only)", align_top_n);
	}
	fprintf(stderr, "...\n");
	
	int i, j, k, l;
	int p = num_clusters/10;
	alignment* al;
	int aligned=0;
	
	int nmer = 4;
	int len = 1;
	int word_num, slot;
	for (i = 0; i < nmer; i++) {	//5^nmers
		len *= 5;
	}
	char **align_matches = calloc(len, sizeof(char*));
	for (i = 0; i < len; i++) {
		align_matches[i] = NULL;
	}
	
	for (i = 0; i < num_barcode_sequences; i++) {
		word_num = strlen(barcode_sequences->entries[i]->key) - nmer + 1;
		for (j = 0; j < word_num; j++) {
			slot = hash_slot_nmer(&barcode_sequences->entries[i]->key[j], nmer);
			if (align_matches[slot] == NULL) {
				align_matches[slot] = calloc(num_barcode_sequences, sizeof(char));
				for (k = 0; k < num_barcode_sequences; k++) {
					align_matches[slot][k] = 0;
				}
			}
			align_matches[slot][i] = 1;	//could ++ it maybe
		}
	}
	
	//word_match is indexed by barcode, representing how many words
	// the currently considered cluster sequence has in common with
	// each barcode
	int* word_match = calloc(num_barcode_sequences, sizeof(int));
	//top_aligns is indexed by cluster, listing each of the barcode
	// sequences aligning maximally to the cluster
	list** top_aligns = calloc(num_clusters, sizeof(list*));
	//best_align_score is indexed by cluster, indicating the maximal
	// alignment score found
	int* best_align_score = calloc(num_clusters, sizeof(int));
	//top_match and top_index are used for tracking the best n candidates
	int* top_match = NULL;
	int* top_index = NULL;
	if (align_top_n > 0) {
		top_match = calloc(align_top_n, sizeof(int));
		top_index = calloc(align_top_n, sizeof(int));
	}
	
	//average_best_align_score for statistics
	int average_best_align_score = 0;
	for (i = 0; i < num_clusters; i++) {
		if (i % p == 0) {
			fprintf(stderr, "\t%d of %d entries [%d%%]\n", i,
					num_clusters, (i*10)/p);
		}
		if (match[i][0] >= 0 || cluster_totals[
				cluster_sequences->entries[i]->value] < align_count_minimum) {
			//only align unmatched entries, skip matched
			//also skip items that don't make the count cutoff
			continue;
		}
		
		top_aligns[i] = list_create();
		best_align_score[i] = -999999;
		
		if (align_top_n > 0) {
			//only align the top n candidates if
			// the user requests it
			for (j = 0; j < num_barcode_sequences; j++) {
				word_match[j] = 0;
			}
			word_num = strlen(cluster_sequences->entries[i]->key) - nmer + 1;
			for (j = 0; j < word_num; j++) {
				slot = hash_slot_nmer(&cluster_sequences->entries[i]->key[j],
						nmer);
				if (align_matches[slot]) {
					for (k = 0; k < num_barcode_sequences; k++) {
						if (align_matches[slot][k]) {	//redundant, but faster
							word_match[k] += align_matches[slot][k];
						}
					}
				}
			}
			
			//all the barcode nmer align_matches have been found
			for (j = 0; j < align_top_n; j++) {
				top_match[j] = 0;
				top_index[j] = -1;
			}
			for (j = 0; j < num_barcode_sequences; j++) {
				for (k = align_top_n-1; k >= 0; k--) {
					if (word_match[j] > top_match[k]) {
						break;
					}
				}
				if (k < 0) {
					continue;
				}
				for (l = 0; l < k; l++) {
					top_match[l] = top_match[l+1];
					top_index[l] = top_index[l+1];
				}
				top_match[k] = word_match[j];
				top_index[k] = j;
			}
			
			for (j = 0; j < align_top_n; j++) {
				if (top_index[j] >= 0) {
					al = align(barcode_sequences->entries[top_index[j]]->key,
							cluster_sequences->entries[i]->key, -1, 5, -1);
					if (al->score >= best_align_score[i]) {
						if (al->score > best_align_score[i]) {
							//discard our old maximal list, since we
							// found a higher score now
							list_destroy(top_aligns[i]);
							top_aligns[i] = list_create();
							
							best_align_score[i] = al->score;
						}
						//add this barcode to the maximal aligning list
						list_set_value(top_aligns[i],
								barcode_sequences->entries[top_index[j]]->key,
								top_index[j]);
					}
					alignment_destroy(al);
				}
			}
		} else {
			//have to attempt an align against everything...
			for (j = 0; j < num_barcode_sequences; j++) {
				al = align(barcode_sequences->entries[j]->key,
						cluster_sequences->entries[i]->key, -1, 5, -1);
				if (al->score >= best_align_score[i]) {
					if (al->score > best_align_score[i]) {
						//discard our old maximal list, since we
						// found a higher score now
						list_destroy(top_aligns[i]);
						top_aligns[i] = list_create();
						
						best_align_score[i] = al->score;
					}
					//add this barcode to the maximal aligning list
					list_set_value(top_aligns[i],
							barcode_sequences->entries[j]->key, j);
				}
				alignment_destroy(al);
			}
		}
		
		aligned++;
		average_best_align_score += best_align_score[i];
	}
	if (aligned > 0) {
		average_best_align_score = average_best_align_score / aligned;
	}
	
	char *tmp;	//will be created and given to match_alignment
	int tmp_len;
	char buff[1024];
	for (i = 0; i < num_clusters; i++) {
		if (match[i][0] >= 0 || cluster_totals[
				cluster_sequences->entries[i]->value] < align_count_minimum) {
			//only align unmatched entries, skip matched
			//also skip items that don't make the count cutoff
			continue;
		}
		if (top_aligns[i]->size > 1) {
			//multiple alignments found
			match_type[i] = MATCH_ALIGNMULTIPLE;
			free(match[i]);	//get rid of the old match list
			match[i] = calloc(top_aligns[i]->size+1, sizeof(int));
			for (j = 0; j < top_aligns[i]->size; j++) {
				match[i][j] = top_aligns[i]->entries[j]->value;
			}
			match[i][top_aligns[i]->size] = -1;
		} else if (top_aligns[i]->size == 1) {
			match_type[i] = MATCH_ALIGN;
			match[i][0] = top_aligns[i]->entries[0]->value;
		} else {
			continue;	//empty list
		}
		
		//copy over the alignment
		
		if (top_aligns[i]->size > 1) {
			//multiple alignments found
			int ij;
			char small_buff[1024];	//for each individual alignment
			buff[0] = '\0';
			buff[1023] = '\0';
			for (j = 0; j < top_aligns[i]->size; j++) {
				ij = top_aligns[i]->entries[j]->value;
				al = align(barcode_sequences->entries[ij]->key,
						cluster_sequences->entries[i]->key, -1, 5, -1);
				sprintf(small_buff, "%d:%s:%d:%d:%d%s", al->score,
								al->seq2, al->mismatches,
								al->inserts, al->deletes,
								(j < top_aligns[i]->size-1 ? ";" : ""));
				strcat(buff, small_buff);
				alignment_destroy(al);
			}
			
			tmp_len = strlen(buff);
			tmp = malloc(tmp_len * (sizeof(char)+1));
			strcpy(tmp, buff);
			tmp[tmp_len] = '\0';
		} else {
			al = align(barcode_sequences->entries[match[i][0]]->key,
					cluster_sequences->entries[i]->key, -1, 5, -1);
			sprintf(buff, "%d:%s:%d:%d:%d", al->score,
					al->seq2, al->mismatches, al->inserts, al->deletes);
			
			tmp_len = strlen(buff);
			tmp = malloc(tmp_len * (sizeof(char)+1));
			strcpy(tmp, buff);
			tmp[tmp_len] = '\0';
			alignment_destroy(al);
		}
		
		//keep track of the alignment information
		// for when we output to the file
		match_alignment[i] = tmp;
	}
	
	free(word_match);
	for (i = 0; i < num_clusters; i++) {
		if (top_aligns[i]) {
			list_destroy(top_aligns[i]);
		}
	}
	free(top_aligns);
	free(best_align_score);
	if (align_top_n > 0) {
		free(top_match);
		free(top_index);
	}
}

/**
 * Entry point for the map method
 * Assigns cluster cores to barcode sequences in two parts
 * If the cluster sequence is within two mismatches of a barcode sequence,
 * it is assigned to it (and the number of mismatches away is noted)
 * The remaining cluster sequences are aligned to the barcode sequences
 */ 
int map_main(int argc, char **argv) {
	if (argc < 2) {
		print_usage_message(argv[0]);
		exit(1);
	} else if (argc > 1 && (strstr(argv[1], "help") || strcmp(argv[1], "?") == 0)) {
		print_help_message(argv[0]);
		exit(0);
	}
	
	int c;
	while ((c = getopt(argc, argv, "c:n:o:s")) != EOF) {
		switch(c) {
		case 'c':
			align_count_minimum = atoi(optarg);
			break;
		case 'n':
			align_top_n = atoi(optarg);
			break;
		case 'o':
			f_out = fopen(optarg, "w");
			if (!f_out) {
				fprintf(stderr, "Could not open %s for writing", optarg);
				exit(1);
			}
			break;
		case 's':
			skip_alignment = 1;
			break;
		default:
			fprintf(stderr, "%s: Unknown parameter (%s)\n", argv[0], optarg);
			exit(1);
		}
	}
	
	if (optind > argc - 1) {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	if (optind == argc - 1) {
		f_cluster = stdin;
		f_barcode = fopen(argv[optind], "r");
	} else if (optind == argc - 2) {
		f_cluster = fopen(argv[optind], "r");
		f_barcode = fopen(argv[optind+1], "r");
	} else {
		print_usage_message(argv[0]);
		exit(1);
	}
	if (f_cluster == NULL) {
		fprintf(stderr, "No cluster input file\n");
		exit(1);
	}
	if (f_barcode == NULL) {
		fprintf(stderr, "Could not open barcode file ");
		if (optind == argc - 1) {
			fprintf(stderr, "%s", argv[optind]);
		} else {
			fprintf(stderr, "%s", argv[optind+1]);
		}
		fprintf(stderr, "\n");
		exit(1);
	}
	
	fprintf(stderr, "Building data from files...\n");
	
	barcode_sequences = list_create();
	barcode_sequences->sorted = 0;
	build_barcode_sequences();
	words = hash_create();
	build_words();
	cluster_sequences = list_create();
	cluster_sequences->sorted = 0;
	build_clusters();
	
	fprintf(stderr, "Performing mismatch matches...\n");
	match = calloc(num_clusters, sizeof(int*));
	match_type = calloc(num_clusters, sizeof(char));
	match_alignment = calloc(num_clusters, sizeof(char*));
	
	int i, j, j2, k, k2, len;
	char *seq, oldchr, oldchr2;
	for (i = 0; i < num_clusters; i++) {
		match[i] = malloc(sizeof(int));
		match[i][0] = -1;
		match_type[i] = MATCH_NONE;
	}
	
	for (i = 0; i < num_clusters; i++) {
		seq = cluster_sequences->entries[i]->key;
		if (strlen(seq) > strlen(barcode_sequences->entries[0]->key)) {
			fprintf(stderr, "\tSkipping (clusters longer than barcodes)\n");
			break;
		}
		match[i][0] = hash_lookup(words, seq);
		if (match[i][0] >= 0) {
			match_type[i] = MATCH_CONTAIN;
			continue;	//found a match, try the next cluster
		}
		//try to see if mismatch is inside
		len = strlen(seq);
		for (j = 0; j < len; j++) {
			oldchr = seq[j];
			for (k = 0; k < 5; k++) {
				seq[j] = "ACGTN"[k];
				if (seq[j] == oldchr) {
					continue;
				}
				match[i][0] = hash_lookup(words, seq);
				if (match[i][0] >= 0) {
					break;
				}
			}
			seq[j] = oldchr;
			if (match[i][0] >= 0) {
				break;
			}
		}
		if (match[i][0] >= 0) {
			match_type[i] = MATCH_MM1;
			continue;	//found a match, try the next cluster
		}
		//try to see if mm2 is inside (SLOW)
		// this is kept at the end in hopes that a
		// match has been found already
		for (j = 0; j < len; j++) {
			oldchr = seq[j];
			for (k = 0; k < 5; k++) {
				seq[j] = "ACGTN"[k];
				if (seq[j] == oldchr) {
					continue;
				}
				//this is the mm1 level
				for (j2 = j; j2 < len; j2++) {
					oldchr2 = seq[j2];
					for (k2 = 0; k2 < 5; k2++) {
						seq[j2] = "ACGTN"[k2];
						if (seq[j2] == oldchr2) {
							continue;
						}
						//this is the mm2 level
						
						match[i][0] = hash_lookup(words, seq);
						if (match[i][0] >= 0) {
							break;
						}
					}
					seq[j2] = oldchr2;
					if (match[i][0] >= 0) {
						break;
					}
				}
				if (match[i][0] >= 0) {
					break;
				}
			}
			seq[j] = oldchr;
			if (match[i][0] >= 0) {
				break;
			}
		}
		if (match[i][0] >= 0) {
			match_type[i] = MATCH_MM2;
			continue;
		}
	}
	
	//align the rest to the barcode sequences
	if (skip_alignment) {
		fprintf(stderr, "Skipping alignments\n");
	} else {
		align_against_barcode_sequences();
	}
	
	//take some statistics
	int contains=0, mm0s=0, mm1s=0, mm2s=0, aligns=0, nones=0;
	for (i = 0; i < num_clusters; i++) {
		if (match_type[i] == MATCH_NONE) {
			nones++;
		} else if (match_type[i] == MATCH_CONTAIN) {
			mm0s++;
		} else if (match_type[i] == MATCH_MM1) {
			mm1s++;
		} else if (match_type[i] == MATCH_MM2) {
			mm2s++;
		} else if (match_type[i] == MATCH_ALIGN ||
				match_type[i] == MATCH_ALIGNMULTIPLE) {
			aligns++;
		}
	}
	contains = mm0s + mm1s + mm2s;
	
	//output results
	if (f_out) {
		fprintf(f_out, "#contains: %d\n#\tmm0s: %d\n#\tmm1s: %d\n#\tmm2s: %d\n" \
				"#aligns: %d\n#misses: %d\n",
				contains, mm0s, mm1s, mm2s, aligns, nones);
		fprintf(f_out, "#core seq\tcount\tscore\tmm0\tmm1n\tmm1\tmm2n\tmm2\tmatch\t" \
				"id\tseq\t[score:alignment:mis:ins:del]...\n");
		for (i = 0; i < num_clusters; i++) {
			j = cluster_sequences->entries[i]->value;
			fprintf(f_out, "%s\t%d\t%d:%d\t%d\t%d\t%d\t%d\t%d\t",
					cluster_sequences->entries[i]->key, cluster_totals[j],
					cluster_mm1n[j], cluster_mm2n[j], cluster_total_mm0[j],
					cluster_mm1n[j], cluster_total_mm1[j],
					cluster_mm2n[j], cluster_total_mm2[j]);
			if (match_type[i] == MATCH_NONE) {
				fprintf(f_out, "none");
			} else if (match_type[i] == MATCH_CONTAIN) {
				fprintf(f_out, "mm0\t");
			} else if (match_type[i] == MATCH_MM1) {
				fprintf(f_out, "mm1\t");
			} else if (match_type[i] == MATCH_MM2) {
				fprintf(f_out, "mm2\t");
			} else if (match_type[i] == MATCH_ALIGN) {
				fprintf(f_out, "align\t");
			} else if (match_type[i] == MATCH_ALIGNMULTIPLE) {
				fprintf(f_out, "align\t");	//can make it specify it here
			} else {
				fprintf(f_out, "N/A\t");
			}
			
			if (match_type[i] != MATCH_ALIGNMULTIPLE) {
				if (match[i][0] >= 0) {
					j = barcode_sequences->entries[match[i][0]]->value;
					fprintf(f_out, "%s\t%s", barcode_details[j],
							barcode_sequences->entries[match[i][0]]->key);
				}
			} else {
				fprintf(f_out, "MULTIPLE:");
				j = 0;
				while (match[i][j] >= 0) {
					fprintf(f_out, "%s", barcode_details[
							barcode_sequences->entries[match[i][j]]->value]);
					if (match[i][j+1] >= 0) {
						fprintf(f_out, ";");
					}
					j++;
				}
				fprintf(f_out, "\t");
				j = 0;
				while (match[i][j] >= 0) {
					fprintf(f_out, "%s", barcode_sequences->entries[match[i][j]]->key);
					if (match[i][j+1] >= 0) {
						fprintf(f_out, ";");
					}
					j++;
				}
			}
			
			if (match_type[i] == MATCH_ALIGN ||
					match_type[i] == MATCH_ALIGNMULTIPLE) {
				fprintf(f_out, "\t%s", match_alignment[i]);
			}
			fprintf(f_out, "\n");
		}
	}
	
	printf("#contains: %d\n#\tmm0s: %d\n#\tmm1s: %d\n#\tmm2s: %d\n" \
			"#aligns: %d\n#misses: %d\n",
			contains, mm0s, mm1s, mm2s, aligns, nones);
	printf("#core seq\tcount\tscore\tmm0\tmm1n\tmm1\tmm2n\tmm2\tmatch\t" \
			"id\tseq\t[score:alignment:mis:ins:del]...\n");
	for (i = 0; i < num_clusters; i++) {
		j = cluster_sequences->entries[i]->value;
		printf("%s\t%d\t%d:%d\t%d\t%d\t%d\t%d\t%d\t",
				cluster_sequences->entries[i]->key, cluster_totals[j],
				cluster_mm1n[j], cluster_mm2n[j], cluster_total_mm0[j],
				cluster_mm1n[j], cluster_total_mm1[j],
				cluster_mm2n[j], cluster_total_mm2[j]);
		if (match_type[i] == MATCH_NONE) {
			printf("none");
		} else if (match_type[i] == MATCH_CONTAIN) {
			printf("mm0\t");
		} else if (match_type[i] == MATCH_MM1) {
			printf("mm1\t");
		} else if (match_type[i] == MATCH_MM2) {
			printf("mm2\t");
		} else if (match_type[i] == MATCH_ALIGN) {
			printf("align\t");
		} else if (match_type[i] == MATCH_ALIGNMULTIPLE) {
			printf("align\t");	//can make it specify it here
		} else {
			printf("N/A\t");
		}
		
		if (match_type[i] != MATCH_ALIGNMULTIPLE) {
			if (match[i][0] >= 0) {
				j = barcode_sequences->entries[match[i][0]]->value;
				printf("%s\t%s", barcode_details[j],
						barcode_sequences->entries[match[i][0]]->key);
			}
		} else {
			printf("MULTIPLE:");
			j = 0;
			while (match[i][j] >= 0) {
				printf("%s", barcode_details[
						barcode_sequences->entries[match[i][j]]->value]);
				if (match[i][j+1] >= 0) {
					printf(";");
				}
				j++;
			}
			printf("\t");
			j = 0;
			while (match[i][j] >= 0) {
				printf("%s", barcode_sequences->entries[match[i][j]]->key);
				if (match[i][j+1] >= 0) {
					printf(";");
				}
				j++;
			}
		}
		
		if (match_type[i] == MATCH_ALIGN ||
				match_type[i] == MATCH_ALIGNMULTIPLE) {
			printf("\t%s", match_alignment[i]);
		}
		printf("\n");
	}
	
	return 0;
}
