#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "hash.h"
#include "list.h"

/**
 * Custom comparator function for use with the built-in qsort
 * Compares the value field of the entry, making larger come before
 * smaller
 */
int entry_comparator(const void* a1, const void* a2) {
	entry* e1 = (entry*) a1;
	entry* e2 = (entry*) a2;
	if (e1->value > e2->value) {
		return -1;
	} else if (e1->value < e2->value) {
		return 1;
	} else {
		return 0;
	}
};

/**
 * Another custom comparator function for use with the built-in qsort
 * Compares value fields of double indirect entries, making large
 * come before small
 */
int indirect_entry_comparator(const void* a1, const void* a2) {
	entry** e1 = (entry**) a1;
	entry** e2 = (entry**) a2;
	if ((*e1)->value > (*e2)->value) {
		return -1;
	} else if ((*e1)->value  < (*e2)->value) {
		return 1;
	} else {
		return 0;
	}
};

/**
 * Sorts a list in place with largest value in front
 * Sets to list to be "unsorted" (by key) after complete
 * Uses the built-in qsort function
 */
void list_sort_by_value(list* l) {
	qsort(l->entries, l->size, sizeof(entry*), indirect_entry_comparator);
	l->sorted = 0;
};


/**
 * Another custom comparator function for use with the built-in qsort
 * Compares key fields of double indirect entries, by strcmp ordering
 */
int indirect_entry_key_comparator(const void* a1, const void* a2) {
	entry** e1 = (entry**) a1;
	entry** e2 = (entry**) a2;
	return strcmp((*e1)->key, (*e2)->key);
};

/**
 * Sorts a list in place by key, to restore a list to sorted order
 * Uses the built-in qsort function
 */
void list_sort_by_key(list* l) {
	qsort(l->entries, l->size, sizeof(entry*), indirect_entry_key_comparator);
	l->sorted = 1;
};

/**
 * Returns an array of entries that are sorted by the value field,
 * largest in front
 * Uses the built-in qsort function to perform the sort quickly
 */
entry* hash_sorted_entries(hash* h) {
	//First just throw all the entries into a single array
	int h_size = hash_entries(h);
	entry* ret = calloc(h_size, sizeof(entry));
	
	int entry_index = 0;
	long i;
	int j;
	char *tmp;
	for (i = 0; i < HASH_SLOTS; i++) {
		for (j = 0; j < h->slots[i]->size; j++) {
			tmp = malloc((strlen(h->slots[i]->entries[j]->key)+1)*sizeof(char));
			strcpy(tmp, h->slots[i]->entries[j]->key);
			tmp[strlen(h->slots[i]->entries[j]->key)] = '\0';
			ret[entry_index].key = tmp;
			ret[entry_index].value = h->slots[i]->entries[j]->value;
			entry_index++;
		}
	}
	
	//Then sort that whole array
	qsort(ret, h_size, sizeof(entry), entry_comparator);
	
	return ret;
};
