#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "include.h"
#include "hash.h"

static hash *barcode_link = NULL;
static char **barcode_details = NULL;
static char **barcode_sequences = NULL;
static int num_barcode_sequences = 0;
static int* barcode_totals = NULL;
static int* barcode_mm0s = NULL;
static int* barcode_mm1s = NULL;
static int* barcode_mm2s = NULL;
static int* barcode_aligns = NULL;
static char** barcode_align_seq = NULL;
FILE* f_barcode = NULL;
FILE* f_map = NULL;
FILE* f_out = NULL;

static int min_align_cutoff = 70;

static void print_usage_message(char *argv0) {
	fprintf(stderr, "Usage: %s [-c align_cutoff] [-o output_file] " \
			"[map_file] barcode_file\n", argv0);
}

static void print_help_message(char *argv0) {
	printf("Usage: %s [map_file] barcode_file\n", argv0);
	printf("Matches the mapped clusters to the barcodes of interest\n");
	printf("map_file specifies the mapped clusters; if unspecified, " \
			"will read from stdin\n");
	printf("barcode_file specifies the barcodes of interest\n\n");
	printf("  -c N\tAlign score needs to be at least N to match a barcode, " \
			"default is %d\n", min_align_cutoff);
}

/**
 * Creates the barcode hash and details from the supplied barcode file
 */
static void build_barcode_links() {
	//assumes f_barcode is valid and barcode_sequences list is ready
	char buff[1024];	buff[1023] = '\0';
	char *ptr, *ptr2;
	while (fgets(buff, 1023, f_barcode)) {
		if (buff[0] == '#') {
			continue;	//commented line; skip it
		}
		
		ptr = strchr(buff, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (sequence\tdetails)\n");
			continue;
		}
		*ptr = '\0';
		ptr++;
		
		ptr2 = strchr(ptr, '\n');
		if (ptr2) {
			*ptr2 = '\0';
		}
		
		hash_set_value(barcode_link, ptr, num_barcode_sequences);
		num_barcode_sequences++;
	}
	
	barcode_details = calloc(num_barcode_sequences, sizeof(char*));
	barcode_sequences = calloc(num_barcode_sequences, sizeof(char*));
	rewind(f_barcode);
	
	char *tmp;
	int tmp_len, i=0;
	while (fgets(buff, 1023, f_barcode)) {
		if (buff[0] == '#') {
			continue;	//commented line; skip it
		}
		
		ptr = strchr(buff, '\t');
		if (ptr == NULL) {
			fprintf(stderr, "Warning: an entry is not in"\
				" expected format (sequence\tdetails)\n");
			continue;
		}
		*ptr = '\0';
		ptr++;
		
		ptr2 = strchr(ptr, '\n');
		if (ptr2) {
			*ptr2 = '\0';
		}
		
		tmp_len = strlen(buff);
		tmp = malloc(tmp_len * (sizeof(char)+1));
		strcpy(tmp, buff);
		tmp[tmp_len] = '\0';
		
		barcode_details[i] = tmp;
		
		tmp_len = strlen(ptr);
		tmp = malloc(tmp_len * (sizeof(char)+1));
		strcpy(tmp, ptr);
		tmp[tmp_len] = '\0';
		
		barcode_sequences[i] = tmp;
		
		i++;
	}
	
	if (num_barcode_sequences != hash_entries(barcode_link)) {
		fprintf(stderr, "Warning: there are non-unique " \
				"barcode_sequences in the barcode file\n");
		num_barcode_sequences = hash_entries(barcode_link);
	}
}

/**
 * Modular function which checks the formatting of a map entry at
 * every field
 */
static void map_format_check_and_process(char **ptr) {
	if (*ptr == NULL) {
		fprintf(stderr, "Error: a map entry is not " \
				"in the valid format:\n\t" \
				"seq\ttotal\tscore\tmm0\tmm1n\tmm1\tmm2n\tmm2\t" \
				"match type\tbarcode id\tbarcode seq\taligns\n");
		exit(1);
	}
	**ptr = '\0';
	(*ptr)++;
}

/**
 * Processes a line from the map file, extracting the information and
 * assigning its total to a matching barcode
 */
static void process_map_line(char *line) {
	char *total_ptr, *mm0_ptr, *mm1_ptr, *mm2_ptr, *matchtype_ptr,
			*barcode_ptr, *align_ptr, *ptr, *ptr2, *ptr3;
	int i;
	
	ptr = strchr(line, '\n');
	if (ptr) {
		*ptr = '\0';
	}
	
	total_ptr = strchr(line, '\t');
	map_format_check_and_process(&total_ptr);
	
	mm0_ptr = strchr(total_ptr, '\t');
	map_format_check_and_process(&mm0_ptr);
	mm0_ptr = strchr(mm0_ptr, '\t');
	map_format_check_and_process(&mm0_ptr);
	
	mm1_ptr = strchr(mm0_ptr, '\t');
	map_format_check_and_process(&mm1_ptr);
	mm1_ptr = strchr(mm1_ptr, '\t');
	map_format_check_and_process(&mm1_ptr);
	
	mm2_ptr = strchr(mm1_ptr, '\t');
	map_format_check_and_process(&mm2_ptr);
	mm2_ptr = strchr(mm2_ptr, '\t');
	map_format_check_and_process(&mm2_ptr);
	
	matchtype_ptr = strchr(mm2_ptr, '\t');
	map_format_check_and_process(&matchtype_ptr);
	
	if (strcmp(matchtype_ptr, "none") == 0) {
		//nothing was matched, so nothing to do here
		return;
	}
	
	barcode_ptr = strchr(matchtype_ptr, '\t');
	map_format_check_and_process(&barcode_ptr);
	barcode_ptr = strchr(barcode_ptr, '\t');
	map_format_check_and_process(&barcode_ptr);
	
	align_ptr = barcode_ptr;
	
	if (strcmp(matchtype_ptr, "align") == 0) {
		align_ptr = strchr(barcode_ptr, '\t');
		map_format_check_and_process(&align_ptr);
		
		//barcode may say MULTIPLE:
		ptr = strchr(barcode_ptr, ':');
		if (ptr) {
			barcode_ptr = ptr+1;
		}
	}
	
	if (strcmp(matchtype_ptr, "align")) {
		//not align, so just write and go
		i = hash_lookup(barcode_link, barcode_ptr);
		if (i >= 0) {
			if (strcmp(matchtype_ptr, "mm0") == 0) {
				barcode_mm0s[i] += atoi(total_ptr);
			} else if (strcmp(matchtype_ptr, "mm1") == 0) {
				barcode_mm1s[i] += atoi(total_ptr);
			} else if (strcmp(matchtype_ptr, "mm2") == 0) {
				barcode_mm2s[i] += atoi(total_ptr);
			}
		}
		return;
	}
	
	int score;
	char buff[8192];	buff[8191] = '\0';
	char *tmp;
	int tmp_len;
	//this is now specifically for aligning
	while (barcode_ptr && align_ptr) {
		//go through entry by entry for multiple aligns
		ptr = strchr(barcode_ptr, ';');
		if (ptr) {
			*ptr = '\0';
			ptr++;
		}
		ptr2 = strchr(align_ptr, ';');
		if (ptr2) {
			*ptr2 = '\0';
			ptr2++;
		}
		
		ptr3 = strchr(align_ptr, ':');
		*ptr3 = '\0';
		score = atoi(align_ptr);
		*ptr3 = ':';
		
		i = hash_lookup(barcode_link, barcode_ptr);
		if (i >= 0 && score >= min_align_cutoff) {
			barcode_aligns[i] += atoi(total_ptr);
			
			if (barcode_align_seq[i]) {
				//was already something there
				strcpy(buff, barcode_align_seq[i]);
				strcat(buff, ";");
				free(barcode_align_seq[i]);
			} else {
				buff[0] = '\0';
			}
			
			strcat(buff, align_ptr);
			tmp_len = strlen(buff);
			tmp = malloc(sizeof(char)*(tmp_len+1));
			strcpy(tmp, buff);
			tmp[tmp_len] = '\0';
			
			barcode_align_seq[i] = tmp;
		}
		
		barcode_ptr = ptr;
		align_ptr = ptr2;
	}
}

/**
 * Entry point for tabulate
 * Takes the barcode file and map entries and tabulates the totals
 */
int tabulate_main(int argc, char **argv) {
	if (argc < 2) {
		print_usage_message(argv[0]);
		exit(1);
	} else if (argc > 1 && (strstr(argv[1], "help") || strcmp(argv[1], "?") == 0)) {
		print_help_message(argv[0]);
		exit(0);
	}
	
	int c;
	while ((c = getopt(argc, argv, "c:n:o:")) != EOF) {
		switch(c) {
		case 'c':
			min_align_cutoff = atoi(optarg);
			break;
		case 'o':
			f_out = fopen(optarg, "w");
			if (!f_out) {
				fprintf(stderr, "Could not open %s for writing", optarg);
				exit(1);
			}
			break;
		default:
			fprintf(stderr, "%s: Unknown parameter (%s)\n", argv[0], optarg);
			exit(1);
		}
	}
	
	if (optind > argc - 1) {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	if (optind == argc - 1) {
		f_map = stdin;
		f_barcode = fopen(argv[optind], "r");
	} else if (optind == argc - 2) {
		f_map = fopen(argv[optind], "r");
		f_barcode = fopen(argv[optind+1], "r");
	} else {
		print_usage_message(argv[0]);
		exit(1);
	}
	
	if (f_map == NULL) {
		fprintf(stderr, "No map file supplied\n");
		exit(1);
	}
	if (f_barcode == NULL) {
		fprintf(stderr, "Could not open barcode file ");
		fprintf(stderr, "%s\n",
				argv[(optind == argc - 1 ? optind : optind + 1)]);
		exit(1);
	}
	
	fprintf(stderr, "Building barcodes...\n");
	
	//Create the barcode hash (barcodes of interest)
	barcode_link = hash_create();
	build_barcode_links();
	
	//Create the variables for holding the results
	barcode_totals = calloc(num_barcode_sequences, sizeof(int));
	barcode_mm0s = calloc(num_barcode_sequences, sizeof(int));
	barcode_mm1s = calloc(num_barcode_sequences, sizeof(int));
	barcode_mm2s = calloc(num_barcode_sequences, sizeof(int));
	barcode_aligns = calloc(num_barcode_sequences, sizeof(int));
	barcode_align_seq = calloc(num_barcode_sequences, sizeof(char*));
	
	int i, j;
	for (i = 0; i < num_barcode_sequences; i++) {
		barcode_totals[i] = 0;
		barcode_mm0s[i] = 0;
		barcode_mm1s[i] = 0;
		barcode_mm2s[i] = 0;
		barcode_aligns[i] = 0;
		barcode_align_seq[i] = NULL;
	}
	
	fprintf(stderr, "Assigning entries...\n");
	char buff[16384];	buff[16383] = '\0';
	//Process all the clusters, line by line
	while (fgets(buff, 16383, f_map)) {
		if (buff[0] == '#') {
			continue;	//skip comments
		}
		
		process_map_line(buff);
	}
	
	//Sort barcodes by alphabetical order
	list* sorted_barcode_ids = list_create();
	sorted_barcode_ids->sorted = 0;
	for (i = 0; i < num_barcode_sequences; i++) {
		barcode_totals[i] = barcode_mm0s[i] + barcode_mm1s[i] +
				barcode_mm2s[i] + barcode_aligns[i];
		list_set_value(sorted_barcode_ids, barcode_details[i], i);
	}
	list_sort_by_key(sorted_barcode_ids);
	
	if (f_out) {
		fprintf(f_out, "#bcid\tbarcode\ttotal\tmm0\tmm1\tmm2\talign\n");
		for (j = 0; j < num_barcode_sequences; j++) {
			i = sorted_barcode_ids->entries[j]->value;
			
			fprintf(f_out, "%s\t%s\t%d\t%d\t%d\t%d\t%d", barcode_details[i],
					barcode_sequences[i], barcode_totals[i],
					barcode_mm0s[i], barcode_mm1s[i], barcode_mm2s[i],
					barcode_aligns[i]);
			if (barcode_align_seq[i]) {
				fprintf(f_out, "\t%s", barcode_align_seq[i]);
			}
			fprintf(f_out, "\n");
		}
	}
	
	printf("#bcid\tbarcode\ttotal\tmm0\tmm1\tmm2\talign\n");
	for (j = 0; j < num_barcode_sequences; j++) {
		i = sorted_barcode_ids->entries[j]->value;
		
		printf("%s\t%s\t%d\t%d\t%d\t%d\t%d", barcode_details[i],
				barcode_sequences[i], barcode_totals[i],
				barcode_mm0s[i], barcode_mm1s[i], barcode_mm2s[i],
				barcode_aligns[i]);
		if (barcode_align_seq[i]) {
			printf("\t%s", barcode_align_seq[i]);
		}
		printf("\n");
	}
	
	return 0;
};
